from typing import Optional
import torch
import torch.nn as nn
import torch.distributed as dist


class LayerCentering(nn.Module):
    r"""
    Applies Layer centering over a mini-batch of inputs.

    This layer implements the operation as described in
    .. math::
        y = x - \mathrm{E}[x] + \beta
    The mean is calculated over the last `D` dimensions
    given in the `dim` parameter.
    `\beta` is learnable bias parameter. that can be
    applied after the mean subtraction.
    Unlike Layer Normalization, this layer is 1-Lipschitz
    This layer uses statistics computed from input data in
    both training and  evaluation modes.

    Args:
        size: number of features in the input tensor
        dim: dimensions over which to compute the mean
        (default ``input.mean((-2, -1))`` for a 4D tensor).
        bias: if `True`, adds a learnable bias to the output
        of shape (size,). Default: `True`

    Shape:
        - Input: :math:`(N, size, *)`
        - Output: :math:`(N, size, *)` (same shape as input)

    """

    def __init__(self, size: int = 1, dim: tuple = [-2, -1], bias=True):
        super(LayerCentering, self).__init__()
        if bias:
            self.bias = nn.Parameter(torch.zeros((size,)), requires_grad=True)
        else:
            self.register_parameter("bias", None)
        self.dim = dim

    def forward(self, x):
        mean = x.mean(dim=self.dim, keepdim=True)
        if self.bias is not None:
            bias_shape = (1, -1) + (1,) * (len(x.shape) - 2)
            return x - mean + self.bias.view(bias_shape)
        else:
            return x - mean


LayerCentering2d = LayerCentering


class BatchCentering(nn.Module):
    r"""
    Applies Batch Centering  over a 2D, 3D, 4D input.

    .. math::

        y = x - \mathrm{E}[x] + \beta

    The mean is calculated per-dimension over the mini-batchesa and
    other dimensions excepted the feature/channel dimension.
    This layer uses statistics computed from input data in
    training mode and  a constant in evaluation mode computed as
    the running mean on training samples.
    :math:`\beta` is a learnable parameter vectors
    of size `C` (where `C` is the number of features or channels of the input).
    that can be applied after the mean subtraction.
    Unlike Batch Normalization, this layer is 1-Lipschitz

    Args:
        size: number of features in the input tensor
        dim: dimensions over which to compute the mean
        (default ``input.mean((0, -2, -1))`` for a 4D tensor).
        momentum: the value used for the running mean computation
        bias: if `True`, adds a learnable bias to the output
        of shape (size,). Default: `True`

    Shape:
        - Input: :math:`(N, size, *)`
        - Output: :math:`(N, size, *)` (same shape as input)

    """

    def __init__(
        self,
        size: int = 1,
        dim: Optional[tuple] = None,
        momentum: float = 0.05,
        bias: bool = True,
    ):
        super(BatchCentering, self).__init__()
        self.dim = dim
        self.momentum = momentum
        self.register_buffer("running_mean", torch.zeros((size,)))
        if bias:
            self.bias = nn.Parameter(torch.zeros((size,)), requires_grad=True)
        else:
            self.register_parameter("bias", None)

        self.first = True

    def forward(self, x):
        if self.dim is None:  # (0,2,3) for 4D tensor; (0,) for 2D tensor
            self.dim = (0,) + tuple(range(2, len(x.shape)))
        mean_shape = (1, -1) + (1,) * (len(x.shape) - 2)
        if self.training:
            mean = x.mean(dim=self.dim)
            with torch.no_grad():
                if self.first:
                    self.running_mean = mean
                    self.first = False
                else:
                    self.running_mean = (
                        1 - self.momentum
                    ) * self.running_mean + self.momentum * mean
            if dist.is_initialized():
                dist.all_reduce(self.running_mean, op=dist.ReduceOp.SUM)
                self.running_mean /= dist.get_world_size()
        else:
            mean = self.running_mean
        if self.bias is not None:
            return x - mean.view(mean_shape) + self.bias.view(mean_shape)
        else:
            return x - mean.view(mean_shape)


BatchCentering2d = BatchCentering



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================

import pytest

import numpy as np

from . import utils_framework as uft
from tests.utils_framework import Lorth2d, LorthRegularizer


@pytest.mark.skipif(
    hasattr(Lorth2d, "unavailable_class"),
    reason="Lorth2d not available",
)
@pytest.mark.parametrize(
    "kernel_shape,stride,delta,padding",
    [
        (None, None, None, None),
        ((5, 5, 32, 64), None, 32, 4),  # RO case
        ((5, 5, 8, 70), 3, 0, 3),  # CO case
    ],
)
def test_set_kernel_shape(kernel_shape, stride, delta, padding):
    """Assert that set_kernel_shape() correctly sets parameters."""
    if stride is None:
        lorth = uft.get_instance_framework(Lorth2d, {})
    else:
        lorth = uft.get_instance_framework(Lorth2d, {"stride": stride})
    if kernel_shape is not None:
        lorth.set_kernel_shape(kernel_shape)
    assert lorth.kernel_shape == kernel_shape
    assert lorth.delta == delta
    assert lorth.padding == padding

    # self.assertIsNone(lorth.kernel_shape)
    # self.assertIsNone(lorth.delta)
    # self.assertIsNone(lorth.padding)

    # kernel_shape = (5, 5, 32, 64)
    # lorth.set_kernel_shape(kernel_shape)
    # self.assertEqual(lorth.kernel_shape, kernel_shape)
    # self.assertEqual(lorth.delta, 32)
    # self.assertEqual(lorth.padding, 4)

    # # CO case
    # lorth = Lorth2d(stride=3)
    # kernel_shape = (5, 5, 8, 70)
    # lorth.set_kernel_shape(kernel_shape)

    # self.assertEqual(lorth.kernel_shape, kernel_shape)
    # self.assertEqual(lorth.delta, 0)  # zero, because CO case
    # self.assertEqual(lorth.padding, 3)


@pytest.mark.skipif(
    hasattr(Lorth2d, "unavailable_class"),
    reason="Lorth2d not available",
)
@pytest.mark.parametrize(
    "kernel_shape,stride,err,err_msg",
    [
        ((3, 3, 32, 289), 4, RuntimeError, "Impossible RO"),
        ((3, 3, 8, 129), 4, RuntimeError, "Impossible CO"),
        ((3, 3, 32, 32), 1, Warning, "hard to optimize"),
        ((4, 4, 32, 32), 1, AssertionError, "odd kernel only"),
    ],
)
def test_existence_orthogonal_conv(kernel_shape, stride, err, err_msg):
    """Assert that Error and Warning are correctly raised when checking existence
    of the orthogonal convolution."""

    # RO case
    if err != Warning:
        with pytest.raises(err):
            uft.get_instance_framework(
                Lorth2d, {"kernel_shape": kernel_shape, "stride": stride}
            )
    else:
        with pytest.warns(err):
            uft.get_instance_framework(
                Lorth2d, {"kernel_shape": kernel_shape, "stride": stride}
            )
        # # CO case
        # with self.assertRaisesRegex(RuntimeError, "Impossible CO"):
        #     kernel_shape = (3, 3, 8, 129)
        #     Lorth2d(kernel_shape, stride=4)

        # # square case (RO=CO)
        # with self.assertWarnsRegex(Warning, "hard to optimize"):
        #     kernel_shape = (3, 3, 32, 32)
        #     Lorth2d(kernel_shape, stride=1)


@pytest.mark.skipif(
    hasattr(Lorth2d, "unavailable_class"),
    reason="Lorth2d not available",
)
def test_compute_lorth():
    """Assert Lorth2d computation on an identity kernel => must return 0"""

    def _identity_kernel(kernel_size, channels):
        assert kernel_size & 1
        kernel_shape = (kernel_size, kernel_size, channels, channels)
        w = np.zeros(kernel_shape, dtype=np.float32)
        for i in range(channels):
            w[kernel_size // 2, kernel_size // 2, i, i] = 1
        return w  # tf.constant(w)

    kernel_size = (np.random.randint(1, 10) * 2) + 1
    channels = np.random.randint(1, 100)
    w = _identity_kernel(kernel_size, channels)

    w = uft.to_tensor(w)
    lorth = uft.get_instance_framework(Lorth2d, {"kernel_shape": w.shape})
    loss = lorth.compute_lorth(w)
    np.testing.assert_allclose(loss.numpy(), 0, atol=1e-6)
    # self.assertAlmostEqual(loss.numpy(), 0)


"""
class TestLorthRegularizer(TestCase):
"""
"""Tests on LorthRegularizer"""


@pytest.mark.skipif(
    hasattr(LorthRegularizer, "unavailable_class"),
    reason="LorthRegularizer not available",
)
def test_LorthRegularizer_init():
    """Check initialization from scratch and from config."""

    kshape1 = (3, 3, 16, 32)
    kshape2 = (5, 5, 16, 8)
    stride = np.random.randint(1, 4)
    lambda_lorth = np.random.normal()
    regul = uft.get_instance_framework(
        LorthRegularizer,
        {"kernel_shape": kshape1, "stride": stride, "lambda_lorth": lambda_lorth},
    )
    # LorthRegularizer(kshape1, stride, lambda_lorth)

    assert regul.lorth.kernel_shape == kshape1
    # self.assertEqual(regul.lorth.kernel_shape, kshape1)

    # Set new shape and get config
    regul.set_kernel_shape(kshape2)
    config = regul.get_config()
    new_regul = LorthRegularizer.from_config(config)

    assert new_regul.lorth.kernel_shape == kshape2
    assert new_regul.stride == stride
    assert new_regul.lambda_lorth == lambda_lorth


@pytest.mark.skipif(
    hasattr(LorthRegularizer, "unavailable_class"),
    reason="LorthRegularizer not available",
)
def test_1D_not_supported():
    """Assert that 1D convolutions are not supported."""
    with pytest.raises(NotImplementedError):
        uft.get_instance_framework(LorthRegularizer, {"dim": 1})



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
This module contains equivalents for Model and Sequential. These classes add support
for condensation and vanilla exportation.
"""

import abc
import copy
import warnings
import math
from collections import OrderedDict
from typing import Any

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize
from torch.nn import Sequential as TorchSequential


def _is_supported_1lip_layer(layer):
    """Return True if the layer is 1-Lipschitz. Note that in some cases, the layer
    is 1-Lipschitz for specific set of parameters.
    """
    supported_1lip_layers = (
        torch.nn.Softmax,
        torch.nn.Flatten,
        torch.nn.Identity,
        torch.nn.ReLU,
        torch.nn.Sigmoid,
        torch.nn.Tanh,
        torch.nn.Unflatten,
    )
    if isinstance(layer, supported_1lip_layers):
        return True
    elif isinstance(layer, torch.nn.MaxPool2d):
        return layer.kernel_size <= layer.stride
    return False


def _get_vanilla_module(module: nn.Module) -> nn.Module:
    """Return the vanilla torch module corresponding to a lipschitz module.

    Args:
        module (nn.Module): A lipschitz module

    Returns:
        nn.Module: The corresponding vanilla torch module
    """
    if isinstance(module, torch.nn.Conv2d):
        return torch.nn.Conv2d(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=module.bias is not None,
            padding_mode=module.padding_mode,
        )
    elif isinstance(module, torch.nn.Linear):
        return torch.nn.Linear(
            in_features=module.in_features,
            out_features=module.out_features,
            bias=module.bias is not None,
        )
    elif isinstance(module, torch.nn.ConvTranspose2d):
        return torch.nn.ConvTranspose2d(
            in_channels=module.in_channels,
            out_channels=module.out_channels,
            kernel_size=module.kernel_size,
            stride=module.stride,
            padding=module.padding,
            output_padding=module.output_padding,
            dilation=module.dilation,
            groups=module.groups,
            bias=module.bias is not None,
            padding_mode=module.padding_mode,
        )
    return module


def vanilla_model(model: nn.Module, in_place=True, new_module=None) -> nn.Module:
    """Convert lipschitz modules into their non-lipschitz counterpart (for
    instance, SpectralConv2d layers become Conv2d layers).

    Warning: This function modifies the model in-place by default.
        Set `in_place=False` for a copy of the model.

    Args:
        model (nn.Module): Lipschitz neural network
        in_place (bool): If True, modify the model in place. If False, return a copy
        new_module (nn.Module): If provided, the function will fill this module
            instead of creating a new one. Only useful in recursion, should be None
            when calling the function.
    """
    if in_place is False and new_module is None:
        model.eval()
        new_module = copy.deepcopy(model)
        new_module.eval()
    if in_place is True:
        new_module = model

    for n, module in model.named_children():
        if isinstance(module, LipschitzModule):
            # torchlip modules
            setattr(new_module, n, module.vanilla_export())
        elif parametrize.is_parametrized(module) and isinstance(
            module, (nn.Conv2d, nn.Linear, nn.ConvTranspose2d)
        ):
            # compatibility with orthogonium modules
            nmod = _get_vanilla_module(module)
            nmod.weight.data = module.weight.data.clone()
            nmod.bias.data = (
                module.bias.data.clone() if module.bias is not None else None
            )
            setattr(new_module, n, nmod)
        elif len(list(module.children())) > 0:
            # compound module, go inside it
            _ = vanilla_model(
                module, in_place=in_place, new_module=new_module.__getattr__(n)
            )
    return new_module


class _LipschitzCoefMultiplication(nn.Module):
    """Parametrization module for lipschitz global coefficient multiplication."""

    def __init__(self, coef: float):
        super().__init__()
        self._coef = coef

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        return self._coef * weight


class LipschitzModule(abc.ABC):
    """
    This class allow to set lipschitz factor of a layer. Lipschitz layer must inherit
    this class to allow user to set the lipschitz factor.

    Warning:
         This class only regroup useful functions when developing new Lipschitz layers.
         But it does not ensure any property about the layer. This means that
         inheriting from this class won't ensure anything about the lipschitz constant.
    """

    # The target coefficient:
    _coefficient_lip: float

    def __init__(self, coefficient_lip: float = 1.0):
        self._coefficient_lip = coefficient_lip

    def apply_lipschitz_factor(self):
        """Multiply the layer weights by a lipschitz factor."""
        if self._coefficient_lip == 1.0:
            return
        parametrize.register_parametrization(
            self, "weight", _LipschitzCoefMultiplication(self._coefficient_lip)
        )

    @abc.abstractmethod
    def vanilla_export(self):
        """
        Convert this layer to a corresponding vanilla torch layer (when possible).

        Returns:
             A vanilla torch version of this layer.
        """
        pass


class Sequential(TorchSequential, LipschitzModule):
    def __init__(
        self,
        *args: Any,
        k_coef_lip: float = 1.0,
    ):
        """
        Equivalent of torch.Sequential but allow to set k-lip factor globally. Also
        support condensation and vanilla exportation.
        For now constant repartition is implemented (each layer
        get n_sqrt(k_lip_factor), where n is the number of layers)
        But in the future other repartition function may be implemented.

        Args:
            layers: list of layers to add to the model.
            name: name of the model, can be None
            k_coef_lip: the Lipschitz coefficient to ensure globally on the model.
        """
        TorchSequential.__init__(self, *args)
        LipschitzModule.__init__(self, k_coef_lip)

        # Force the Lipschitz coefficient:
        n_layers = np.sum(
            [isinstance(layer, LipschitzModule) for layer in self.children()]
        )
        for module in self.children():
            if isinstance(module, LipschitzModule):
                module._coefficient_lip = math.pow(k_coef_lip, 1 / n_layers)
            elif _is_supported_1lip_layer(module) is not True:
                warnings.warn(
                    "Sequential model contains a layer which is not a Lipschitz layer: {}".format(  # noqa: E501
                        module
                    )
                )

    def vanilla_export(self):
        """
        Exports this model to a vanilla torch Sequential.

        This method only works for flat sequential. Lipschitz modules are converted
        using their own `vanilla_export` method while non-Lipschitz modules are simply
        copied using `copy.deepcopy`.

        Returns:
            A Vanilla torch.nn.Sequential model.
        """
        layers = []
        for name, layer in self.named_children():
            if isinstance(layer, LipschitzModule):
                layers.append((name, layer.vanilla_export()))
            else:
                layers.append((name, copy.deepcopy(layer)))
        return TorchSequential(OrderedDict(layers))



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
This module contains computation functions for Bjorck and spectral
normalization. This is most for internal use.
"""

from typing import Optional
from typing import Tuple

import torch
import torch.nn.functional as F

DEFAULT_EPS_SPECTRAL = 1e-3
DEFAULT_EPS_BJORCK = 1e-3
DEFAULT_MAXITER_BJORCK = 15
DEFAULT_MAXITER_SPECTRAL = 10
DEFAULT_BETA = 0.5
DEFAULT_BJORK_DETACH = 4


def bjorck_normalization(
    w: torch.Tensor,
    eps: float = DEFAULT_EPS_BJORCK,
    beta: float = DEFAULT_BETA,
    maxiter: int = DEFAULT_MAXITER_BJORCK,
) -> torch.Tensor:
    r"""
    Apply Bjorck normalization on the kernel as per

    .. math::
        \begin{array}{l}
            W_0 = W \\
            W_{n + 1} = (1 + \beta) W_{n} - \beta W_n W_n^T W_n \\
            \overline{W} = W_{N}
        \end{array}

    where :math:`W` is the kernel of shape :math:`(C, *)`, with :math:`C` the number
    of channels.

    Args:
        w: Weights to normalize. For the normalization to work properly, the greatest
            eigen value of ``w`` must be approximately 1.
        eps (float): epsilon stopping criterion: norm(wt - wt-1) must be less than eps
        beta (float): beta used in each iteration, must be in the interval ]0, 0.5]
        maxiter (int): maximum number of iterations for the algorithm

    Returns:
        The weights :math:`\overline{W}` after Bjorck normalization.
    """

    def cond(w, old_w):
        return torch.linalg.norm(w - old_w) >= eps

    # define the loop body
    def body_cols(w):
        return torch.mm(w, torch.mm(w.t(), w))

    def body_rows(w):
        return torch.mm(torch.mm(w, w.t()), w)

    def body(w, fct, iter):
        if maxiter - iter >= DEFAULT_BJORK_DETACH:
            return (1.0 + beta) * w - beta * fct(w).detach()
        w = (1.0 + beta) * w - beta * fct(w)
        return w

    shape = w.shape
    cout = w.size(0)
    w_mat = w.reshape(cout, -1)

    if w_mat.shape[0] > w_mat.shape[1]:
        body_fct = body_cols
    else:
        body_fct = body_rows

    done = False
    iter = maxiter

    while not done:
        old_w = w_mat
        w_mat = body(w_mat, body_fct, iter)
        iter -= 1
        done = (not cond(w_mat, old_w)) or (iter <= 0)

    w = w_mat.reshape(shape)
    return w


def _power_iteration(
    linear_operator,
    adjoint_operator,
    u: torch.Tensor,
    eps=DEFAULT_EPS_SPECTRAL,
    maxiter=DEFAULT_MAXITER_SPECTRAL,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Internal function that performs the power iteration algorithm.

    Args:
        linear_operator (Callable): a callable object that maps a linear operation.
        adjoint_operator (Callable): a callable object that maps the adjoint of the
            linear operator.
        u (tf.Tensor): initialization of the singular vector.
        eps (float, optional): stopping criterion of the algorithm, when
            norm(u[t] - u[t-1]) is less than eps. Defaults to DEFAULT_EPS_SPECTRAL.
        maxiter (int, optional): maximum number of iterations for the algorithm.
            Defaults to DEFAULT_MAXITER_SPECTRAL.

    Returns:
         A Tensor containing the maximum singular vector

    """

    # Loop stopping condition
    def cond(u, old_u):
        return torch.linalg.norm(u - old_u) >= eps

    # Loop body
    def body(u):
        v = linear_operator(u)
        u = adjoint_operator(v)

        u = F.normalize(u, dim=-1)

        return u

    done = False
    iter = maxiter

    while not done:
        old_u = u
        u = body(u)
        iter -= 1
        done = (not cond(u, old_u)) or (iter <= 0)

    return u


def spectral_normalization(
    kernel: torch.Tensor,
    u: Optional[torch.Tensor] = None,
    eps: float = DEFAULT_EPS_SPECTRAL,
    maxiter: int = DEFAULT_MAXITER_SPECTRAL,
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    r"""
    Apply spectral normalization on the given kernel :math:`W` to set its greatest
    eigen value to approximately 1.

    .. note::
        This function is provided for completeness since ``torchlip`` layers use
        the :py:func:`torch.nn.utils.spectral_norm` hook instead.

    Args:
        kernel: The kernel to normalize.
        u: Initialization for the initial singular vector. If ``None``, it will
            be randomly initialized from a normal distribution and twice
            as much iterations will be performed.
        eps (float, optional): stopping criterion of the algorithm, when
            norm(u[t] - u[t-1]) is less than eps. Defaults to DEFAULT_EPS_SPECTRAL.
        maxiter (int, optional): maximum number of iterations for the algorithm.
            Defaults to DEFAULT_MAXITER_SPECTRAL.

    Returns:
        The normalized kernel :math:`\overline{W}`, the right singular vector
        corresponding to the largest singular value, and the largest singular value
        before normalization.
    """

    def linear_op(u):
        return u @ kernel.t()

    def adjoint_op(v):
        return v @ kernel

    # Flatten the Tensor
    W_flat = kernel.flatten(start_dim=1)

    if u is None:
        maxiter *= 2  # if u was not double number of iterations for the first run
        u = torch.ones(tuple([1, W_flat.shape[-1]]), device=kernel.device)
        torch.nn.init.normal_(u)

    # do power iteration
    u = _power_iteration(linear_op, adjoint_op, u, eps, maxiter)

    # Compute the largest singular value and the normalized kernel.
    # We assume that in the worst case we converged to sigma + eps (as u and v are
    # normalized after each iteration)
    # In order to be sure that operator norm of normalized kernel is strictly less than
    # one we use sigma + eps, which ensures stability of Björck algorithm even when
    # beta=0.5
    sigma = torch.linalg.norm(linear_op(u))
    normalized_kernel = kernel.div(sigma + eps)
    return normalized_kernel, u, sigma



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import os
import pprint
import pytest
import numpy as np
from .test_layers import linear_generator, build_kernel

from . import utils_framework as uft

from tests.utils_framework import (
    Sequential,
    tModel,
)

from tests.utils_framework import (
    SpectralLinear,
    SpectralConv2d,
    SpectralConvTranspose2d,
    FrobeniusLinear,
    FrobeniusConv2d,
    ScaledL2NormPool2d,
)
from tests.utils_framework import Flatten, tLinear, tInput

np.random.seed(42)
pp = pprint.PrettyPrinter(indent=4)


def sequential_layers(input_shape):
    """Return list of layers for a Sequential model"""
    in_ch = input_shape[0]
    input_shape = uft.to_framework_channel(input_shape)
    return [
        uft.get_instance_framework(tInput, {"shape": input_shape}),
        uft.get_instance_framework(
            SpectralConv2d,
            {
                "in_channels": in_ch,
                "out_channels": 2,
                "kernel_size": (3, 3),
                "padding": 1,
            },
        ),
        uft.get_instance_framework(ScaledL2NormPool2d, {"kernel_size": (2, 2)}),
        uft.get_instance_framework(
            FrobeniusConv2d,
            {"in_channels": 2, "out_channels": 2, "kernel_size": (3, 3), "padding": 1},
        ),
        uft.get_instance_framework(
            SpectralConvTranspose2d,
            {"in_channels": 2, "out_channels": 5, "kernel_size": (3, 3), "padding": 1},
        ),
        uft.get_instance_framework(Flatten, {}),
        uft.get_instance_framework(tLinear, {"in_features": 80, "out_features": 4}),
        uft.get_instance_framework(
            SpectralLinear, {"in_features": 4, "out_features": 4}
        ),
        uft.get_instance_framework(
            FrobeniusLinear, {"in_features": 4, "out_features": 2}
        ),
    ]


def get_functional_tensors(input_shape):
    in_ch = input_shape[0]
    input_shape = uft.to_framework_channel(input_shape)
    dict_functional_tensors = {}
    dict_functional_tensors["inputs"] = uft.get_instance_framework(
        tInput, {"shape": input_shape}
    )
    dict_functional_tensors["conv1"] = uft.get_instance_framework(
        SpectralConv2d,
        {
            "in_channels": in_ch,
            "out_channels": 2,
            "kernel_size": (3, 3),
            "k_coef_lip": 2.0,
            "padding": 1,
        },
    )
    dict_functional_tensors["pool1"] = uft.get_instance_framework(
        ScaledL2NormPool2d, {"kernel_size": (2, 2), "k_coef_lip": 2.0}
    )
    dict_functional_tensors["conv2"] = uft.get_instance_framework(
        FrobeniusConv2d,
        {
            "in_channels": 2,
            "out_channels": 2,
            "kernel_size": (3, 3),
            "k_coef_lip": 2.0,
            "padding": 1,
        },
    )
    dict_functional_tensors["convt2"] = uft.get_instance_framework(
        SpectralConvTranspose2d,
        {"in_channels": 2, "out_channels": 5, "kernel_size": (3, 3), "padding": 1},
    )
    dict_functional_tensors["flatten"] = uft.get_instance_framework(Flatten, {})
    dict_functional_tensors["dense1"] = uft.get_instance_framework(
        tLinear, {"in_features": 80, "out_features": 4}
    )
    dict_functional_tensors["dense2"] = uft.get_instance_framework(
        SpectralLinear, {"in_features": 4, "out_features": 4}
    )
    dict_functional_tensors["dense3"] = uft.get_instance_framework(
        FrobeniusLinear, {"in_features": 4, "out_features": 2}
    )
    return dict_functional_tensors


def functional_input_output_tensors(dict_functional_tensors, x):
    """Return input and output tensor of a Functional (hard-coded) model"""
    if dict_functional_tensors["inputs"] is None:
        inputs = x
    else:
        inputs = dict_functional_tensors["inputs"]
    x = dict_functional_tensors["conv1"](inputs)
    x = dict_functional_tensors["pool1"](x)
    x = dict_functional_tensors["conv2"](x)
    x = dict_functional_tensors["convt2"](x)
    x = dict_functional_tensors["flatten"](x)
    x = dict_functional_tensors["dense1"](x)
    x = dict_functional_tensors["dense2"](x)
    outputs = dict_functional_tensors["dense3"](x)
    if dict_functional_tensors["inputs"] is None:
        return outputs
    else:
        return inputs, outputs
    # return x


def get_model(model_type, layer_params, input_shape, k_coef_lip):
    if model_type == tModel:
        return uft.get_functional_model(
            tModel,
            layer_params["dict_tensors"],
            layer_params["functional_input_output_tensors"],
        )
    else:
        return uft.generate_k_lip_model(
            model_type, layer_params, input_shape=input_shape, k=k_coef_lip
        )


@pytest.mark.skipif(
    hasattr(SpectralConvTranspose2d, "unavailable_class"),
    reason="SpectralConvTranspose2d not available",
)
@pytest.mark.parametrize(
    "model_type, params_type, param_fct, dict_other_params, k_coef_lip, input_shape",
    [
        (Sequential, "layers", sequential_layers, {}, 5.0, (3, 8, 8)),
        (
            tModel,
            "dict_tensors",
            get_functional_tensors,
            {
                "functional_input_output_tensors": functional_input_output_tensors,
            },
            5.0,
            (3, 8, 8),
        ),
    ],
)
def test_model(
    model_type, params_type, param_fct, dict_other_params, k_coef_lip, input_shape
):
    batch_size = 250
    epochs = 1
    steps_per_epoch = 125
    k_lip_data = 2.0

    # clear session to avoid side effects from previous train
    uft.init_session()  # K.clear_session()
    np.random.seed(42)
    input_shape_CHW = input_shape
    input_shape = uft.to_framework_channel(input_shape)
    layer_params = {params_type: param_fct(input_shape_CHW)}
    layer_params.update(dict_other_params)
    model = get_model(model_type, layer_params, input_shape, k_coef_lip)

    # create the model, defin opt, and compile it
    optimizer = uft.get_instance_framework(
        uft.Adam, inst_params={"lr": 0.001, "model": model}
    )

    loss_fn, optimizer, metrics = uft.compile_model(
        model,
        optimizer=optimizer,
        loss=uft.MeanSquaredError(),
        metrics=[uft.metric_mse()],
    )
    # create the synthetic data generator
    output_shape = uft.compute_output_shape(input_shape, model)
    # output_shape = model.uft.compute_output_shape((batch_size,) + input_shape)[1:]
    kernel = build_kernel(input_shape, output_shape, k_lip_data)
    # define logging features
    logdir = os.path.join(
        "logs", uft.LIP_LAYERS, "condense_test"
    )  # , datetime.now().strftime("%Y_%m_%d-%H_%M_%S")))
    os.makedirs(logdir, exist_ok=True)
    callback_list = []  # callbacks.TensorBoard(logdir)]
    # train model

    traind_ds = linear_generator(batch_size, input_shape, kernel)
    uft.train(
        traind_ds,
        model,
        loss_fn,
        optimizer,
        epochs,
        batch_size,
        steps_per_epoch=steps_per_epoch,
        callbacks=callback_list,
    )
    # the seed is set to compare all models with the same data
    np.random.seed(42)
    # get original results
    test_dl = linear_generator(batch_size, input_shape, kernel)
    loss, mse = uft.run_test(model, test_dl, loss_fn, metrics, steps=10)
    # generate vanilla
    if uft.vanilla_require_a_copy():
        layer_params = {params_type: param_fct(input_shape_CHW)}
        layer_params.update(dict_other_params)
        model2 = get_model(model_type, layer_params, input_shape, k_coef_lip)
        uft.copy_model_parameters(model, model2)
        vanilla_model = uft.vanillaModel(model2)
    else:
        vanilla_model = uft.vanillaModel(model)
    # vanilla_model = model.vanilla_export()
    loss_fn, optimizer, metrics = uft.compile_model(
        vanilla_model,
        optimizer=optimizer,
        loss=uft.MeanSquaredError(),
        metrics=[uft.metric_mse()],
    )
    np.random.seed(42)
    # evaluate vanilla
    test_dl = linear_generator(batch_size, input_shape, kernel)
    loss2, mse2 = uft.run_test(model, test_dl, loss_fn, metrics, steps=10)
    np.random.seed(42)
    # check if original has changed
    test_dl = linear_generator(batch_size, input_shape, kernel)
    vanilla_loss, vanilla_mse = uft.run_test(
        vanilla_model, test_dl, loss_fn, metrics, steps=10
    )

    np.testing.assert_almost_equal(
        mse,
        vanilla_mse,
        3,
        "the exported vanilla model must have same behaviour as original",
    )
    np.testing.assert_equal(
        mse, mse2, "exporting a model must not change original model"
    )
    # add one epoch to orginal

    traind_ds = linear_generator(batch_size, input_shape, kernel)
    uft.train(
        traind_ds,
        model,
        loss_fn,
        optimizer,
        epochs,
        batch_size,
        steps_per_epoch=steps_per_epoch,
        callbacks=callback_list,
    )
    np.random.seed(42)
    test_dl = linear_generator(batch_size, input_shape, kernel)
    loss3, mse3 = uft.run_test(model, test_dl, loss_fn, metrics, steps=10)
    # check if vanilla has changed
    np.random.seed(42)
    test_dl = linear_generator(batch_size, input_shape, kernel)
    vanilla_loss2, vanilla_mse2 = uft.run_test(
        vanilla_model, test_dl, loss_fn, metrics, steps=10
    )
    np.testing.assert_almost_equal(
        vanilla_mse,
        vanilla_mse2,
        5,
        "exported model must be completely independent from original",
    )

    with pytest.raises(AssertionError):
        np.testing.assert_almost_equal(
            mse,
            mse3,
            4,
            "all tests passe but integrity check failed: the test cannot conclude that "
            + "vanilla_export create a distinct model",
        )



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import torch
from torch.nn.utils.parametrizations import spectral_norm

from ..utils import bjorck_norm
from ..normalizers import DEFAULT_EPS_BJORCK
from ..normalizers import DEFAULT_EPS_SPECTRAL
from ..utils import frobenius_norm
from .module import LipschitzModule


class SpectralLinear(torch.nn.Linear, LipschitzModule):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        k_coef_lip: float = 1.0,
        eps_spectral: int = DEFAULT_EPS_SPECTRAL,
        eps_bjorck: int = DEFAULT_EPS_BJORCK,
    ):
        """
        This class is a Linear Layer constrained such that all singular of it's kernel
        are 1. The computation based on BjorckNormalizer algorithm.
        The computation is done in two steps:

        1. reduce the larget singular value to 1, using iterated power method.
        2. increase other singular values to 1, using BjorckNormalizer algorithm.

        Args:
            in_features: Size of each input sample.
            out_features: Size of each output sample.
            bias: If ``False``, the layer will not learn an additive bias.
            k_coef_lip: Lipschitz constant to ensure.
            eps_spectral: stopping criterion for the iterative power algorithm.
            eps_bjorck: stopping criterion Bjorck algorithm (DEFAULT_EPS_BJORCK),
             to discard Bjorck projection (1-Lip only) set this parameter to None.

        Shape:
            - Input: :math:`(N, *, H_{in})` where :math:`*` means any number of
              additional dimensions and :math:`H_{in} = \\text{in\\_features}`
            - Output: :math:`(N, *, H_{out})` where all but the last dimension
              are the same shape as the input and
              :math:`H_{out} = \\text{out\\_features}`.

        This documentation reuse the body of the original torch.nn.Linear doc.
        """
        torch.nn.Linear.__init__(
            self,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        spectral_norm(
            self,
            name="weight",
            eps=eps_spectral,
        )
        if eps_bjorck is not None:  # discard the bjorck norm if None
            bjorck_norm(self, name="weight", eps=eps_bjorck)
        self.apply_lipschitz_factor()

    def vanilla_export(self) -> torch.nn.Linear:
        layer = torch.nn.Linear(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=self.bias is not None,
        )
        layer.weight.data = self.weight.detach()
        if self.bias is not None:
            layer.bias.data = self.bias.detach()
        return layer


class FrobeniusLinear(torch.nn.Linear, LipschitzModule):
    """
    This class is a Linear Layer constrained such that the Frobenius norm of the weight
    is 1. In the case of a single output neuron, it is equivalent and faster than the
    SpectralLinear layer. For multi-neuron case, the "disjoint_neurons" parameter
    affects the behaviour:

    - if ``disjoint_neurons`` is True (default), it corresponds to the stacking of
      independent 1-Lipschitz neurons.
    - if ``disjoint_neurons`` is False, the matrix weight is normalized by its Frobenius
      norm.

    Args:
        in_features: Size of each input sample.
        out_features: Size of each output sample.
        bias: If ``False``, the layer will not learn an additive bias.
        disjoint_neurons: Normalize, independently per neuron or not, the matrix weight.
        k_coef_lip: Lipschitz constant to ensure.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        disjoint_neurons: bool = True,
        k_coef_lip: float = 1.0,
    ):
        torch.nn.Linear.__init__(
            self,
            in_features=in_features,
            out_features=out_features,
            bias=bias,
        )
        LipschitzModule.__init__(self, k_coef_lip)

        torch.nn.init.orthogonal_(self.weight)
        if self.bias is not None:
            self.bias.data.fill_(0.0)

        frobenius_norm(self, name="weight", disjoint_neurons=disjoint_neurons)
        self.apply_lipschitz_factor()

    def vanilla_export(self):
        layer = torch.nn.Linear(
            in_features=self.in_features,
            out_features=self.out_features,
            bias=self.bias is not None,
        )
        layer.weight.data = self.weight.detach()
        if self.bias is not None:
            layer.bias.data = self.bias.detach()
        return layer



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Custom autograd function for safe-gradient computation of square-root at 0.
"""

from typing import Any

import torch


class SqrtEpsGrad(torch.autograd.Function):
    """
    Small class to avoid division by zero when computing the gradient
    of the sqrt function.
    """

    @staticmethod
    def forward(ctx: Any, input: Any, eps: float) -> torch.Tensor:  # type: ignore
        sqrt_input = torch.sqrt(input)
        ctx.save_for_backward(sqrt_input)
        ctx.eps = eps
        return sqrt_input

    @staticmethod
    def backward(ctx: Any, grad_output):  # type: ignore
        (input,) = ctx.saved_tensors
        return grad_output / (2 * (input + ctx.eps)), None


def sqrt_with_gradeps(input: torch.Tensor, eps: float = 1e-6) -> torch.Tensor:
    r"""
    Square-root of input with a "valid" gradient at 0.

    .. math::
        \frac{\partial f}{\partial x} = \frac{1}{2\sqrt{x}+\epsilon}

    Args:
        input: Tensor of arbitrary shape.
        eps: Value to add to the input when computing gradient (must be positive).

    Returns:
        A tensor whose value is the square-root of the input but whose associated
        autograd functions is :py:class:`SqrtEpsGrad`.
    """
    return SqrtEpsGrad.apply(input, eps)  # type: ignore



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest
import numpy as np

from . import utils_framework as uft

from .utils_framework import (
    SpectralInitializer,
    tLinear,
)


@pytest.mark.skipif(
    hasattr(SpectralInitializer, "unavailable_class"),
    reason="SpectralInitializer not available",
)
@pytest.mark.parametrize(
    "layer_type, layer_params,input_shape, orthogonal_test",
    [
        (
            tLinear,
            {
                "in_features": 5,
                "out_features": 4,
                "kernel_initializer": uft.get_instance_framework(
                    SpectralInitializer,
                    inst_params={"eps_spectral": 1e-6, "eps_bjorck": 1e-6},
                ),
            },
            (5,),
            True,
        ),
        (
            tLinear,
            {
                "in_features": 5,
                "out_features": 100,
                "kernel_initializer": uft.get_instance_framework(
                    SpectralInitializer,
                    inst_params={"eps_spectral": 1e-6, "eps_bjorck": None},
                ),
            },
            (5,),
            False,
        ),
    ],
)
def test_initializer(layer_type, layer_params, input_shape, orthogonal_test):
    np.random.seed(42)
    # clear session to avoid side effects from previous train
    uft.init_session()  # K.clear_session()
    input_shape = uft.to_framework_channel(input_shape)
    # create the model, defin opt, and compile it
    model = uft.generate_k_lip_model(layer_type, layer_params, input_shape)
    uft.initialize_kernel(model, 0, layer_params["kernel_initializer"])

    optimizer = uft.get_instance_framework(
        uft.Adam, inst_params={"lr": 0.001, "model": model}
    )  # uft.Adam(lr=0.001)

    loss_fn, optimizer, metrics = uft.compile_model(
        model,
        optimizer=optimizer,
        loss=uft.MeanSquaredError(),
        metrics=[uft.metric_mse()],
    )

    sigmas = np.linalg.svd(
        uft.to_numpy(uft.get_layer_weights_by_index(model, 0)),
        full_matrices=False,
        compute_uv=False,
    )
    if orthogonal_test:
        np.testing.assert_allclose(sigmas, np.ones_like(sigmas), atol=1e-5)
    else:
        np.testing.assert_allclose(sigmas.max(), 1.0, atol=2e-2)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import os
import pytest

import numpy as np

from . import utils_framework as uft

from .utils_framework import BatchCentering, LayerCentering


def check_serialization(layer_type, layer_params, input_shape=(10,)):
    m = uft.generate_k_lip_model(layer_type, layer_params, input_shape=input_shape, k=1)
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.CategoricalCrossentropy(from_logits=True),
    )
    name = layer_type.__class__.__name__
    path = os.path.join("logs", "normalization", name)
    xnp = np.random.uniform(-10, 10, (255,) + input_shape)
    x = uft.to_tensor(xnp)
    y1 = m(x)
    uft.save_model(m, path)
    m2 = uft.load_model(
        path,
        compile=True,
        layer_type=layer_type,
        layer_params=layer_params,
        input_shape=input_shape,
        k=1,
    )
    y2 = m2(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2))


@pytest.mark.skipif(
    hasattr(LayerCentering, "unavailable_class"),
    reason="LayerCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_LayerCentering(size, input_shape, bias):
    """evaluate layerbatch centering"""
    input_shape = uft.to_framework_channel(input_shape)
    x = np.arange(np.prod(input_shape)).reshape(input_shape)
    bn = uft.get_instance_framework(LayerCentering, {"size": size, "bias": bias})

    mean_x = np.mean(x, axis=(2, 3))
    mean_shape = (-1, size, 1, 1)
    x = uft.to_tensor(x)
    y = bn(x)
    np.testing.assert_allclose(
        uft.to_numpy(y), x - np.reshape(mean_x, mean_shape), atol=1e-5
    )
    y = bn(2 * x)
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # keep substract batch mean
    bn.eval()
    y = bn(2 * x)
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # eval mode use running_mean


@pytest.mark.skipif(
    hasattr(BatchCentering, "unavailable_class"),
    reason="BatchCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4), False),
        (4, (3, 4), True),
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_BatchCentering(size, input_shape, bias):
    """evaluate layerbatch centering"""
    input_shape = uft.to_framework_channel(input_shape)
    x = np.arange(np.prod(input_shape)).reshape(input_shape)
    bn = uft.get_instance_framework(BatchCentering, {"size": size, "bias": bias})
    bn_mom = bn.momentum
    if len(input_shape) == 2:
        mean_x = np.mean(x, axis=0)
        mean_shape = (1, size)
    else:
        mean_x = np.mean(x, axis=(0, 2, 3))
        mean_shape = (1, size, 1, 1)
    x = uft.to_tensor(x)
    y = bn(x)
    np.testing.assert_allclose(bn.running_mean, mean_x, atol=1e-5)
    np.testing.assert_allclose(
        uft.to_numpy(y), x - np.reshape(mean_x, mean_shape), atol=1e-5
    )
    y = bn(2 * x)
    new_runningmean = mean_x * (1 - bn_mom) + 2 * mean_x * bn_mom
    np.testing.assert_allclose(bn.running_mean, new_runningmean, atol=1e-5)
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - 2 * np.reshape(mean_x, mean_shape), atol=1e-5
    )  # keep substract batch mean
    bn.eval()
    y = bn(2 * x)
    np.testing.assert_allclose(
        bn.running_mean, new_runningmean, atol=1e-5
    )  # eval mode running mean freezed
    np.testing.assert_allclose(
        uft.to_numpy(y), 2 * x - np.reshape(new_runningmean, mean_shape), atol=1e-5
    )  # eval mode use running_mean


@pytest.mark.parametrize(
    "norm_type",
    [LayerCentering, BatchCentering],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), False),
        (10, (10,), True),
        (7, (7, 8, 8), False),
        (7, (7, 8, 8), True),
    ],
)
def test_Normalization_serialization(norm_type, size, input_shape, bias):
    # Check serialization
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    check_serialization(
        norm_type, layer_params={"size": size, "bias": bias}, input_shape=input_shape
    )


def linear_generator(batch_size, input_shape: tuple):
    """
    Generate data according to a linear kernel
    Args:
        batch_size: size of each batch
        input_shape: shape of the desired input

    Returns:
        a generator for the data

    """
    input_shape = tuple(input_shape)
    while True:
        # pick random sample in [0, 1] with the input shape
        batch_x = np.array(
            np.random.uniform(-10, 10, (batch_size,) + input_shape), dtype=np.float16
        )
        # apply the k lip linear transformation
        batch_y = batch_x
        yield batch_x, batch_y


@pytest.mark.parametrize(
    "norm_type",
    [LayerCentering, BatchCentering],
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (10, (10,), True),
        (7, (7, 8, 8), True),
    ],
)
def test_Normalization_bias(norm_type, size, input_shape, bias):
    if hasattr(norm_type, "unavailable_class"):
        pytest.skip(f"{norm_type} not available")
    m = uft.generate_k_lip_model(
        norm_type,
        layer_params={"size": size, "bias": bias},
        input_shape=input_shape,
        k=1,
    )
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.CategoricalCrossentropy(from_logits=True),
    )
    batch_size = 10
    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    np.testing.assert_allclose(bb, np.zeros((size,)), atol=1e-5)

    traind_ds = linear_generator(batch_size, input_shape)
    uft.train(
        traind_ds,
        m,
        loss,
        optimizer,
        2,
        batch_size,
        steps_per_epoch=10,
    )

    bb = uft.to_numpy(uft.get_layer_by_index(m, 0).bias)
    assert np.linalg.norm(bb) != 0.0


@pytest.mark.skipif(
    hasattr(BatchCentering, "unavailable_class"),
    reason="BatchCentering not available",
)
@pytest.mark.parametrize(
    "size, input_shape, bias",
    [
        (4, (3, 4), False),
        (4, (3, 4), True),
        (4, (3, 4, 8, 8), False),
        (4, (3, 4, 8, 8), True),
    ],
)
def test_BatchCentering_runningmean(size, input_shape, bias):
    """evaluate batch centering convergence of running mean"""
    input_shape = uft.to_framework_channel(input_shape)
    # start with 0 to set up running mean to zero
    x = np.zeros(input_shape)
    bn = uft.get_instance_framework(BatchCentering, {"size": size, "bias": bias})
    x = uft.to_tensor(x)
    y = bn(x)

    np.testing.assert_allclose(bn.running_mean, 0.0, atol=1e-5)

    x = np.random.normal(0.0, 1.0, input_shape)
    if len(input_shape) == 2:
        mean_x = np.mean(x, axis=0)
    else:
        mean_x = np.mean(x, axis=(0, 2, 3))
    x = uft.to_tensor(x)
    for _ in range(1000):
        y = bn(x)  # noqa: F841

    np.testing.assert_allclose(bn.running_mean, mean_x, atol=1e-5)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
Contains utility functions.
"""

from .bjorck_norm import bjorck_norm
from .bjorck_norm import remove_bjorck_norm
from .frobenius_norm import frobenius_norm
from .frobenius_norm import remove_frobenius_norm
from .lconv_norm import lconv_norm
from .lconv_norm import remove_lconv_norm
from .sqrt_eps import sqrt_with_gradeps  # noqa: F401
from .evaluate_lip_const import evaluate_lip_const

__all__ = [
    "bjorck_norm",
    "remove_bjorck_norm",
    "frobenius_norm",
    "remove_frobenius_norm",
    "lconv_norm",
    "remove_lconv_norm",
    "sqrt_with_gradeps",
    "evaluate_lip_const",
]



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================

from typing import Optional, Union
import torch

if torch.__version__.startswith("1."):
    import functorch as tfc
else:
    import torch.func as tfc


def evaluate_lip_const(
    model: torch.nn.Module,
    x: Optional[torch.Tensor] = None,  # can be torch tensor or None
    evaluation_type: Union[str, list[str]] = "jacobian_norm",
    disjoint_neurons=False,
    input_shape=None,
    expected_value=None,
    **kwargs,
) -> float:
    """
    Evaluate the Lipschitz constant of a model, using different methods.
    Please note that the estimation of the lipschitz constant is done locally around
    input samples. This may not correctly estimate the behaviour in the whole domain.
    Args:
        model: built torch model used to make predictions
        x: inputs used to compute the lipschitz constant. If None, input_shape must be
            provided to generate random inputs. (shape: (batch_size, ...))
        evaluation_type: method used to evaluate the lipschitz constant. Can be one of
            "jacobian_norm", "noise_norm", "attack" or "all".
        disjoint_neurons: if True, each output neuron is considered as a separate
            1-Lipschitz function.
        input_shape: shape of the input tensor, used if x is None.
        expected_value: if provided, the computed lipschitz constant is compared to
            this value and an assertion error is raised if the computed value is higher
            (warning the assertion is strict).
    """
    type2fct = {
        "jacobian_norm": evaluate_lip_const_jacobian_norm,
        "noise_norm": evaluate_lip_const_noise_norm,
        "attack": evaluate_lip_const_attack,
    }
    device = "cpu"
    if len(list(model.parameters())) > 0:
        device = next(model.parameters()).device

    if evaluation_type == "all":
        evaluation_type = list(type2fct.keys())
    if isinstance(evaluation_type, list):
        lip_csts = []
        for etype in evaluation_type:
            lip_csts.append(
                evaluate_lip_const(
                    model,
                    x,
                    etype,
                    disjoint_neurons,
                    input_shape,
                    expected_value,
                    **kwargs,
                )
            )
        return float(torch.max(torch.tensor(lip_csts)).item())

    if evaluation_type not in type2fct:
        raise ValueError(
            f"Unknown evaluation_type {evaluation_type}. "
            f"Available types are {list(type2fct.keys())}"
        )
    if x is None:
        if input_shape is None:
            raise ValueError("If x is None, input_shape must be provided")
        x = torch.randn(input_shape)
    x = x.to(device)
    val_lip_const = type2fct[evaluation_type](model, x, disjoint_neurons, **kwargs)
    print(
        f"Empirical lipschitz constant is {val_lip_const} with method {evaluation_type}"
    )
    if expected_value is not None:
        assert (
            val_lip_const <= expected_value
        ), f"Empirical lipschitz constant {val_lip_const} is \
            higher than expected value {expected_value}"
    return val_lip_const


def evaluate_lip_const_noise_norm(
    model: torch.nn.Module,
    x: torch.Tensor,
    disjoint_neurons=False,
    num_noisy_samples: int = 10,
    epsilon_noise: float = 1.0,
    **kwargs,
) -> float:
    """
    Evaluate the Lipschitz constant of a model, using random noise added to the input.
    Please note that the estimation of the lipschitz constant is done locally around
    input samples. This may not correctly estimate the behaviour in the whole domain.
    Args:
        model: built torch model used to make predictions
        x: inputs used to compute the lipschitz constant (shape: (batch_size, ...))
        disjoint_neurons: if True, each output neuron is considered as a separate
            1-Lipschitz function.
        num_noisy_samples: number of random noise samples to use for the estimation
        epsilon_noise: standard deviation of the gaussian noise added to the input
    Returns:
        float: the empirically evaluated Lipschitz constant (max over batch).
    """
    with torch.no_grad():
        model.eval()
        pred = model(x)
        lip_csts = []
        for _ in range(num_noisy_samples):  # random sampling
            noise = epsilon_noise * torch.randn_like(x) * torch.rand(1).to(x.device)
            noisy_input = x + noise
            noisy_pred = model(noisy_input)
            if not disjoint_neurons:
                pred_diff_norm = torch.linalg.norm(
                    (pred - noisy_pred).view(pred.shape[0], -1), dim=1
                )
            else:
                # each output neuron is a 1Lipschitz function
                diff_pred = pred - noisy_pred
                diff_pred = diff_pred.view(diff_pred.shape[0], -1, diff_pred.shape[-1])
                pred_diff_norm = torch.linalg.norm(diff_pred, dim=1)
                pred_diff_norm = torch.max(pred_diff_norm, dim=-1).values

            input_diff_norm = torch.linalg.norm(noise.view(pred.shape[0], -1), dim=1)
            lip_cst = pred_diff_norm / input_diff_norm
            lip_csts.append(lip_cst)
        lip_csts = torch.cat(lip_csts, dim=0)
        return float(torch.max(lip_csts).item())


def evaluate_lip_const_jacobian_norm(
    model: torch.nn.Module,
    x: torch.Tensor,
    disjoint_neurons=False,
    **kwargs,
) -> float:
    """
    Evaluate the Lipschitz constant of a model, using the Jacobian of the model.
    Please note that the estimation of the lipschitz constant is done locally around
    input samples. This may not correctly estimate the behaviour in the whole domain.

    Args:
        model: built torch model used to make predictions
        x: inputs used to compute the lipschitz constant
        disjoint_neurons: if True, each output neuron is considered as a separate
            1-Lipschitz function.

    Returns:
        float: the empirically evaluated Lipschitz constant. The computation might also
            be inaccurate in high dimensional space.

    """

    # Define a function that computes the model output
    def model_func(x):
        # using vmap torchfunc method induce a single sample input
        # so we need to unsqueeze the input
        y = model(torch.unsqueeze(x, dim=0))  # Forward pass
        return y

    # assert disjoint_neurons is False, "disjoint_neurons=True not implemented yet"
    x_src = x.clone().detach().requires_grad_(True)

    # Compute the Jacobian using jacrev
    batch_jacobian = tfc.vmap(tfc.jacrev(model_func))(x_src)
    # Reshape the Jacobian to match the desired shape
    batch_size = x.shape[0]
    xdim = torch.prod(torch.tensor(x.shape[1:])).item()

    if not disjoint_neurons:
        batch_jacobian = batch_jacobian.view(batch_size, -1, xdim)
    else:
        # each output neuron is a 1Lipschitz function:
        # compute the norm of each output neuron
        outdim = len(batch_jacobian.shape) - len(x.shape[1:]) - 1
        outsize = batch_jacobian.shape[outdim]
        # switch outdim to be the first dimension
        batch_jacobian = batch_jacobian.moveaxis(outdim, 0)
        batch_jacobian = batch_jacobian.reshape(outsize * batch_size, -1, xdim)

    # Compute singular values and check Lipschitz property
    lip_cst = torch.linalg.norm(batch_jacobian, ord=2, dim=(-2, -1))
    return float(torch.max(lip_cst).item())


def _compute_disjoint_neurons_lip_const(
    ref_output: torch.Tensor, noisy_pred: torch.Tensor
) -> torch.Tensor:
    # each output neuron is a 1Lipschitz function: attack the maximum
    diff_pred = ref_output - noisy_pred
    diff_pred = diff_pred.view(diff_pred.shape[0], diff_pred.shape[1], -1)
    pred_diff_norm = torch.linalg.norm(diff_pred, dim=-1)
    pred_diff_norm = torch.max(pred_diff_norm, dim=1).values
    return pred_diff_norm


def evaluate_lip_const_attack(
    model: torch.nn.Module,
    x: torch.Tensor,
    disjoint_neurons=False,
    num_iterations: int = 100,
    step_size: float = 1e-2,
    double_attack: bool = False,
    **kwargs,
) -> float:
    """
    Evaluate the Lipschitz constant of a model, using an adversarial attack.
    Please note that the estimation of the lipschitz constant is done locally around
    input samples. This may not correctly estimate the behaviour in the whole domain.

    Args:
        model: built torch model used to make predictions
        x: inputs used to compute the lipschitz constant
        disjoint_neurons: if True, each output neuron is considered as a separate
            1-Lipschitz function.
        num_iterations: number of iterations for the attack
        step_size: step size for each iteration
        double_attack: if True, perform a second attack starting from the worst case
            found during the first attack. This can improve the estimation of the
            Lipschitz constant at the cost of doubling the computation time.

    Returns:
        float: the empirically evaluated Lipschitz constant. The computation might also
            be inaccurate in high dimensional space.

    """

    def attack_lip_const(model, x, step_size, num_iterations):
        model.eval()
        with torch.no_grad():
            ref_output = model(x).detach()
        noise = (torch.randn_like(x) * torch.rand(1).to(x.device)).detach()
        for _ in range(num_iterations):
            noise = noise.requires_grad_(True)
            noisy_input = x + noise
            noisy_pred = model(noisy_input)

            if not disjoint_neurons:
                pred_diff_norm = torch.linalg.norm(
                    (ref_output - noisy_pred).view(ref_output.shape[0], -1), dim=1
                )
            else:
                # each output neuron is a 1Lipschitz function: attack the maximum
                pred_diff_norm = _compute_disjoint_neurons_lip_const(
                    ref_output, noisy_pred
                )
            input_diff_norm = torch.linalg.norm(
                noise.view(ref_output.shape[0], -1), dim=1
            )
            lip_cst = pred_diff_norm / input_diff_norm
            loss = lip_cst.mean()

            grad = torch.autograd.grad(loss, noise)[0]
            noise = noise + step_size * grad.sign()
            noise = noise.detach()

        noisy_input = x + noise
        noisy_pred = model(noisy_input)
        if not disjoint_neurons:
            pred_diff_norm = torch.linalg.norm(
                (ref_output - noisy_pred).view(ref_output.shape[0], -1), dim=1
            )
        else:
            # each output neuron is a 1Lipschitz function: attack the maximum
            pred_diff_norm = _compute_disjoint_neurons_lip_const(ref_output, noisy_pred)
        input_diff_norm = torch.linalg.norm(noise.view(ref_output.shape[0], -1), dim=1)
        lip_cst = pred_diff_norm / input_diff_norm
        return noise, lip_cst

    if double_attack:
        print("Warning : double_attack is set to True, \
                the computation time will be doubled")
    noise, lip_cst = attack_lip_const(model, x, step_size, num_iterations)

    if not double_attack:
        return float(torch.max(lip_cst).item())

    # double attack

    noise2, lip_cst2 = attack_lip_const(model, x + noise, step_size, num_iterations)
    return float(torch.max(torch.cat([lip_cst, lip_cst2], dim=0)).item())



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import pytest

import numpy as np

from . import utils_framework as uft

from .utils_framework import pad_input, PadConv2d
from .utils_framework import tConv2d
from .utils_framework import vanillaModel


def compare(x, x_ref, index_x=[], index_x_ref=[]):
    """Compare a tensor and its padded version, based on index_x and ref."""
    x = uft.to_NCHW(uft.to_numpy(x))
    x_ref = uft.to_NCHW(uft.to_numpy(x_ref))
    x_cropped = x[:, :, index_x[0] : index_x[1], index_x[3] : index_x[4]][
        :, :, :: index_x[2], :: index_x[5]
    ]
    if index_x_ref[0] is None:  # compare with 0
        np.testing.assert_allclose(x_cropped, np.zeros(x_cropped.shape), 1e-2, 0)
    else:
        np.testing.assert_allclose(
            x_cropped
            - x_ref[
                :, :, index_x_ref[0] : index_x_ref[1], index_x_ref[3] : index_x_ref[4]
            ][:, :, :: index_x_ref[2], :: index_x_ref[5]],
            np.zeros(x_cropped.shape),
            1e-2,
            0,
        )


@pytest.mark.parametrize(
    "padding_tested", ["circular", "constant", "symmetric", "reflect", "replicate"]
)
@pytest.mark.parametrize(
    "input_shape, batch_size, kernel_size, filters",
    [
        ((1, 5, 5), 250, (3, 3), 2),
        ((5, 5, 5), 250, (3, 3), 2),
        ((2, 7, 7), 250, (7, 7), 2),
        ((128, 5, 5), 250, 3, 2),
    ],
)
def test_padding(padding_tested, input_shape, batch_size, kernel_size, filters):
    """Test different padding types: assert values in original and padded tensors"""
    input_shape = uft.to_framework_channel(input_shape)
    if not uft.is_supported_padding(padding_tested, PadConv2d):
        pytest.skip(f"Padding {padding_tested} not supported")
    kernel_size_list = kernel_size
    if isinstance(kernel_size, (int, float)):
        kernel_size_list = [kernel_size, kernel_size]

    x = np.random.normal(size=(batch_size,) + input_shape).astype("float32")
    x = uft.to_tensor(x)
    x_pad = pad_input(x, padding_tested, kernel_size)
    p_vert, p_hor = kernel_size_list[0] // 2, kernel_size_list[1] // 2
    x_pad_NCHW = uft.get_NCHW(x_pad)
    x_NCHW = uft.get_NCHW(x)

    center_x_pad = [p_vert, -p_vert, 1, p_hor, -p_hor, 1, "center"]
    upper_x_pad = [0, p_vert, 1, p_hor, -p_hor, 1, "upper"]
    lower_x_pad = [-p_vert, x_pad_NCHW[2], 1, p_hor, -p_hor, 1, "lower"]
    left_x_pad = [p_vert, -p_vert, 1, 0, p_hor, 1, "left"]
    right_x_pad = [p_vert, -p_vert, 1, -p_hor, x_pad_NCHW[3], 1, "right"]
    all_x = [0, x_NCHW[2], 1, 0, x_NCHW[3], 1]
    upper_x = [0, p_vert, 1, 0, x_NCHW[3], 1]
    upper_x_first = [0, 1, 1, 0, x_NCHW[3], 1]
    upper_x_rev = [0, p_vert, -1, 0, x_NCHW[3], 1]
    upper_x_refl = [1, p_vert + 1, -1, 0, x_NCHW[3], 1]
    lower_x = [-p_vert, x_NCHW[2], 1, 0, x_NCHW[3], 1]
    lower_x_last = [-1, x_NCHW[2], 1, 0, x_NCHW[3], 1]
    lower_x_rev = [-p_vert, x_NCHW[2], -1, 0, x_NCHW[3], 1]
    lower_x_refl = [-p_vert - 1, x_NCHW[2] - 1, -1, 0, x_NCHW[3], 1]
    left_x = [0, x_NCHW[2], 1, 0, p_hor, 1]
    left_x_first = [0, x_NCHW[2], 1, 0, 1, 1]
    left_x_rev = [0, x_NCHW[2], 1, 0, p_hor, -1]
    left_x_refl = [0, x_NCHW[2], 1, 1, p_hor + 1, -1]
    right_x = [0, x_NCHW[2], 1, -p_hor, x_NCHW[3], 1]
    right_x_last = [0, x_NCHW[2], 1, -1, x_NCHW[3], 1]
    right_x_rev = [0, x_NCHW[2], 1, -p_hor, x_NCHW[3], -1]
    right_x_refl = [0, x_NCHW[2], 1, -p_hor - 1, x_NCHW[3] - 1, -1]
    zero_pad = [None, None, None, None]
    pad_tests = [
        {
            "circular": [center_x_pad, all_x],
            "constant": [center_x_pad, all_x],
            "symmetric": [center_x_pad, all_x],
            "reflect": [center_x_pad, all_x],
            "replicate": [center_x_pad, all_x],
        },
        {
            "circular": [upper_x_pad, lower_x],
            "constant": [upper_x_pad, zero_pad],
            "symmetric": [upper_x_pad, upper_x_rev],
            "reflect": [upper_x_pad, upper_x_refl],
            "replicate": [upper_x_pad, upper_x_first],
        },
        {
            "circular": [lower_x_pad, upper_x],
            "constant": [lower_x_pad, zero_pad],
            "symmetric": [lower_x_pad, lower_x_rev],
            "reflect": [lower_x_pad, lower_x_refl],
            "replicate": [lower_x_pad, lower_x_last],
        },
        {
            "circular": [left_x_pad, right_x],
            "constant": [left_x_pad, zero_pad],
            "symmetric": [left_x_pad, left_x_rev],
            "reflect": [left_x_pad, left_x_refl],
            "replicate": [left_x_pad, left_x_first],
        },
        {
            "circular": [right_x_pad, left_x],
            "constant": [right_x_pad, zero_pad],
            "symmetric": [right_x_pad, right_x_rev],
            "reflect": [right_x_pad, right_x_refl],
            "replicate": [right_x_pad, right_x_last],
        },
    ]

    for test_pad in pad_tests:
        compare(
            x_pad,
            x,
            index_x=test_pad[padding_tested][0],
            index_x_ref=test_pad[padding_tested][1],
        )


@pytest.mark.skipif(
    hasattr(PadConv2d, "unavailable_class"),
    reason="PadConv2d not available",
)
@pytest.mark.parametrize(
    "padding_tested",
    ["circular", "constant", "symmetric", "reflect", "replicate", "same", "valid"],
)
@pytest.mark.parametrize(
    "input_shape, batch_size, kernel_size, filters",
    [
        ((1, 5, 5), 250, (3, 3), 2),
        ((5, 5, 5), 250, (3, 3), 2),
        ((2, 7, 7), 250, (7, 7), 2),
        ((128, 5, 5), 250, 3, 2),
    ],
)
def test_predict(padding_tested, input_shape, batch_size, kernel_size, filters):
    """Compare predictions between pad+Conv2d and PadConv2d layers."""
    in_ch = input_shape[0]
    input_shape = uft.to_framework_channel(input_shape)

    if not uft.is_supported_padding(padding_tested, PadConv2d):
        pytest.skip(f"Padding {padding_tested} not supported")
    layer_params = {
        "out_channels": 2,
        "in_channels": in_ch,
        "kernel_size": (3, 3),
        "bias": False,
        "padding": 0,
        "padding_mode": "zeros",
    }
    layer_params["kernel_size"] = kernel_size
    layer_params["out_channels"] = filters
    if isinstance(kernel_size, (int, float)):
        ks = kernel_size
    else:
        ks = kernel_size[0]
    x = np.random.normal(size=(batch_size,) + input_shape).astype("float32")
    x = uft.to_tensor(x)
    x_pad = pad_input(x, padding_tested, layer_params["kernel_size"])
    layer_params_ref = layer_params.copy()
    if padding_tested.lower() == "same":
        layer_params_ref["padding"] = ks // 2  # same

    model_ref = uft.generate_k_lip_model(
        layer_type=tConv2d,
        layer_params=layer_params_ref,
        input_shape=x_pad.shape[1:],
        k=1.0,
    )
    y_ref = uft.compute_predict(model_ref, x_pad)

    layer_params_pad = layer_params.copy()

    if padding_tested.lower() == "valid":
        layer_params_pad["padding"] = 0
    else:
        layer_params_pad["padding"] = ks // 2
    layer_params_pad["padding_mode"] = padding_tested
    model = uft.generate_k_lip_model(
        layer_type=PadConv2d,
        layer_params=layer_params_pad,
        input_shape=input_shape,
        k=1.0,
    )
    uft.copy_model_parameters(model_ref, model)
    y = uft.compute_predict(model, x)
    y_ref = uft.to_numpy(y_ref)
    y = uft.to_numpy(y)

    np.testing.assert_allclose(y_ref, y, 1e-2, 0)


@pytest.mark.skipif(
    hasattr(PadConv2d, "unavailable_class"),
    reason="PadConv2d not available",
)
@pytest.mark.parametrize(
    "padding_tested",
    ["circular", "constant", "symmetric", "reflect", "replicate", "same", "valid"],
)
@pytest.mark.parametrize(
    "input_shape, batch_size, kernel_size, filters",
    [
        ((1, 5, 5), 250, (3, 3), 2),
        ((5, 5, 5), 250, (3, 3), 2),
        ((2, 7, 7), 250, (7, 7), 2),
        ((128, 5, 5), 250, 3, 2),
    ],
)
def test_vanilla(padding_tested, input_shape, batch_size, kernel_size, filters):
    """Compare predictions between PadConv2d and its vanilla export."""
    in_ch = input_shape[0]
    input_shape = uft.to_framework_channel(input_shape)

    if not uft.is_supported_padding(padding_tested, PadConv2d):
        pytest.skip(f"Padding {padding_tested} not supported")
    layer_params = {
        "out_channels": 2,
        "in_channels": in_ch,
        "kernel_size": (3, 3),
        "bias": False,
        "padding": 0,
        "padding_mode": "zeros",
    }
    layer_params["kernel_size"] = kernel_size
    layer_params["out_channels"] = filters
    if isinstance(kernel_size, (int, float)):
        ks = kernel_size
    else:
        ks = kernel_size[0]
    x = np.random.normal(size=(batch_size,) + input_shape).astype("float32")
    x = uft.to_tensor(x)
    layer_params_pad = layer_params.copy()
    if padding_tested.lower() == "valid":
        layer_params_pad["padding"] = 0
    else:
        layer_params_pad["padding"] = ks // 2
    layer_params_pad["padding_mode"] = padding_tested
    model = uft.generate_k_lip_model(
        layer_type=PadConv2d,
        layer_params=layer_params_pad,
        input_shape=input_shape,
        k=1.0,
    )
    y = uft.compute_predict(model, x)

    if uft.vanilla_require_a_copy():
        model2 = uft.generate_k_lip_model(
            layer_type=PadConv2d,
            layer_params=layer_params_pad,
            input_shape=input_shape,
            k=1.0,
        )

        uft.copy_model_parameters(model, model2)
        model_v = vanillaModel(model2)
    else:
        model_v = vanillaModel(model)
    y_v = uft.compute_predict(model_v, x)
    y_v = uft.to_numpy(y_v)
    y = uft.to_numpy(y)
    np.testing.assert_allclose(y_v, y, 1e-2, 0)



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize

from ..normalizers import bjorck_normalization
from ..normalizers import DEFAULT_EPS_BJORCK


class _BjorckNorm(nn.Module):
    def __init__(self, weight: torch.Tensor, eps: float) -> None:
        super().__init__()
        self.eps = eps
        self.register_buffer("_w_bjorck", weight.data)

    def forward(self, weight: torch.Tensor) -> torch.Tensor:
        if self.training:
            w_bjorck = bjorck_normalization(weight, self.eps)
            self._w_bjorck = w_bjorck.data
        else:
            w_bjorck = self._w_bjorck
        return w_bjorck


def bjorck_norm(
    module: nn.Module, name: str = "weight", eps: float = DEFAULT_EPS_BJORCK
) -> nn.Module:
    r"""
    Applies Bjorck normalization to a parameter in the given module.

    Bjorck normalization ensures that all eigen values of a vectors remain close or
    equal to one during training. If the dimension of the weight tensor is greater than
    2, it is reshaped to 2D for iteration.
    This is implemented via a Bjorck normalization parametrization.

    .. note::
        It is recommended to use :py:func:`torch.nn.utils.parameterize.spectral_norm`
        before this hook to greatly reduce the number of iterations required.

    See `Sorting out Lipschitz function approximation
    <https://arxiv.org/abs/1811.05381>`_.

    Args:
        module: Containing module.
        name: Name of weight parameter.
        eps: stopping criterion Bjorck algorithm.

    Returns:
        The original module with the Bjorck normalization hook.

    Example:

        >>> m = bjorck_norm(nn.Linear(20, 40), name='weight')
        >>> m
        Linear(in_features=20, out_features=40, bias=True)

    See Also:
        :py:func:`deel.torchlip.normalizers.bjorck_normalization`
    """
    weight = getattr(module, name, None)
    parametrize.register_parametrization(module, name, _BjorckNorm(weight, eps))
    return module


def remove_bjorck_norm(module: nn.Module, name: str = "weight") -> nn.Module:
    r"""
    Removes the Bjorck normalization reparameterization from a module.

    Args:
        module: Containing module.
        name: Name of weight parameter.

    Example:
        >>> m = bjorck_norm(nn.Linear(20, 40))
        >>> remove_bjorck_norm(m)
    """
    for key, m in module.parametrizations[name]._modules.items():
        if isinstance(m, _BjorckNorm):
            if len(module.parametrizations["weight"]) == 1:
                parametrize.remove_parametrizations(module, name)
            else:
                del module.parametrizations[name]._modules[key]



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import os
import pytest

import numpy as np

from . import utils_framework as uft

from .utils_framework import LipResidual
from .utils_framework import tInput, tSplit, tModel


def get_functional_tensors(input_shape):
    dict_functional_tensors = {}
    dict_functional_tensors["inputs"] = uft.get_instance_framework(
        tInput, {"shape": input_shape}
    )
    dict_functional_tensors["split"] = uft.get_instance_framework(
        tSplit, {"chunks": 2, "dim": 1}
    )
    dict_functional_tensors["residual"] = uft.get_instance_framework(LipResidual, {})
    return dict_functional_tensors


def functional_input_output_tensors(dict_functional_tensors, x):
    """Return input and output tensor of a Functional (hard-coded) model"""
    if dict_functional_tensors["inputs"] is None:
        inputs = x
    else:
        inputs = dict_functional_tensors["inputs"]
    x = dict_functional_tensors["split"](inputs)
    outputs = dict_functional_tensors["residual"](x[0], x[1])
    if dict_functional_tensors["inputs"] is None:
        return outputs
    else:
        return inputs, outputs
    # return x


def check_serialization(layer_type, layer_params, input_shape=(10,)):
    dict_tensors = get_functional_tensors(input_shape)
    m = uft.get_functional_model(tModel, dict_tensors, functional_input_output_tensors)
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.MeanSquaredError(),
    )
    name = layer_type.__class__.__name__
    path = os.path.join("logs", "residual", name)
    xnp = np.random.uniform(-10, 10, (255,) + input_shape)
    x = uft.to_tensor(xnp)
    y1 = m(x)
    uft.save_model(m, path)

    # build and generate the model
    if uft.vanilla_require_a_copy():
        dict_tensors2 = get_functional_tensors(input_shape)
        m2 = uft.get_functional_model(
            tModel, dict_tensors2, functional_input_output_tensors
        )
        m2 = uft.load_state_dict(path, m2)
    else:
        m2 = uft.load_model(
            path,
            compile=True,
            layer_type=layer_type,
            layer_params=layer_params,
            input_shape=input_shape,
            k=1,
        )
    y2 = m2(x)
    np.testing.assert_allclose(uft.to_numpy(y1), uft.to_numpy(y2))


@pytest.mark.skipif(
    hasattr(LipResidual, "unavailable_class"),
    reason="LipResidual not available",
)
@pytest.mark.parametrize(
    "input_shape",
    [
        (3, 4, 8, 8),
    ],
)
def test_initLipResidual(input_shape):
    """evaluate layerbatch centering"""
    input_shape = uft.to_framework_channel(input_shape)
    x1 = np.arange(np.prod(input_shape)).reshape(input_shape)
    x2 = np.zeros(input_shape)
    res = uft.get_instance_framework(LipResidual, {})

    alpha_res = uft.to_numpy(res.alpha)
    assert alpha_res == 0.0
    z = res(uft.to_tensor(x1), uft.to_tensor(x1))
    np.testing.assert_allclose(uft.to_numpy(z), x1, atol=1e-5)
    z = res(uft.to_tensor(x1), uft.to_tensor(x2))
    np.testing.assert_allclose(uft.to_numpy(z), x1 / 2.0, atol=1e-5)


@pytest.mark.skipif(
    hasattr(LipResidual, "unavailable_class"),
    reason="LipResidual not available",
)
@pytest.mark.parametrize(
    "input_shape",
    [
        (14, 8, 8),
    ],
)
def test_Normalization_serialization(input_shape):
    # Check serialization
    check_serialization(LipResidual, layer_params={}, input_shape=input_shape)


def linear_generator(batch_size, input_shape: tuple, input_type: str):
    """
    Generate data according to a linear kernel
    Args:
        batch_size: size of each batch
        input_shape: shape of the desired input
        input_type: duplication type for residual

    Returns:
        a generator for the data

    """
    input_shape = tuple(
        [sh // 2 if id == 0 else sh for id, sh in enumerate(input_shape)]
    )
    while True:
        # pick random sample in [0, 1] with the input shape
        batch_x = np.array(
            np.random.uniform(-10, 10, (batch_size,) + input_shape), dtype=np.float16
        )
        # same output as input
        batch_y = batch_x
        # concatenate to use split
        if input_type == "zeros":
            batch_x = np.concatenate([batch_x, np.zeros_like(batch_x)], axis=1)
        if input_type == "invert":
            batch_x = np.concatenate([np.zeros_like(batch_x), batch_x], axis=1)
        if input_type == "copy":
            batch_x = np.concatenate([batch_x, batch_x], axis=1)
        if input_type == "random":
            batch_x = np.concatenate(
                [
                    batch_x,
                    np.array(
                        np.random.uniform(-10, 10, (batch_size,) + input_shape),
                        dtype=np.float16,
                    ),
                ],
                axis=1,
            )
        yield batch_x, batch_y


def sigmoid(z):
    return 1 / (1 + np.exp(-z))


@pytest.mark.skipif(
    hasattr(LipResidual, "unavailable_class"),
    reason="LipResidual not available",
)
@pytest.mark.parametrize(
    "input_shape, input_type, learnt_alpha",
    [
        ((14, 8, 8), "zeros", 1.0),  # x1=x x2=0
        ((14, 8, 8), "copy", 0.5),  # x1=x2=x
        ((14, 8, 8), "invert", 0.0),  # x1=0 x2=x
        ((14, 8, 8), "random", None),  # x1=x1 x2=x2
    ],
)
def test_learntResidual(input_shape, input_type, learnt_alpha):
    dict_tensors = get_functional_tensors(input_shape)
    m = uft.get_functional_model(tModel, dict_tensors, functional_input_output_tensors)
    if m is None:
        pytest.skip()
    loss, optimizer, _ = uft.compile_model(
        m,
        optimizer=uft.get_instance_framework(uft.SGD, inst_params={"model": m}),
        loss=uft.MeanSquaredError(),
    )
    batch_size = 10

    traind_ds = linear_generator(batch_size, input_shape, input_type)
    uft.train(
        traind_ds,
        m,
        loss,
        optimizer,
        5,
        batch_size,
        steps_per_epoch=100,
    )

    alpha = uft.to_numpy(m.get_module_by_name("residual").alpha)
    if learnt_alpha is not None:
        np.testing.assert_allclose(sigmoid(alpha), learnt_alpha, atol=1e-1)
    else:
        assert alpha != 0.0



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
"""
This module extends original Pytorch layers, in order to add k-Lipschitz constraint via
re-parametrization. Currently, are implemented:

* Linear layer:
    as SpectralLinear
* Conv1d layer:
    as SpectralConv1d
* Conv2d layer:
    as SpectralConv2d
* Conv3d layer:
    as SpectralConv3d
* AvgPool2d:
    as ScaledAvgPool2d

By default the layers are 1 Lipschitz almost everywhere, which is efficient for
wasserstein distance estimation. However for other problems (such as adversarial
robustness) the user may want to use layers that are at most 1 lipschitz, this can
be done by setting the param `niter_bjorck=0`.
"""

# flake8: noqa
from .activation import FullSort
from .activation import GroupSort
from .activation import GroupSort2
from .activation import HouseHolder
from .activation import LPReLU
from .activation import MaxMin
from .conv import FrobeniusConv2d
from .conv import SpectralConv2d
from .conv import SpectralConv1d
from .conv import SpectralConvTranspose2d
from .downsampling import InvertibleDownSampling
from .linear import FrobeniusLinear
from .linear import SpectralLinear
from .loss import HingeMarginLoss
from .loss import HingeMulticlassLoss
from .loss import HKRLoss
from .loss import HKRMulticlassLoss
from .loss import SoftHKRMulticlassLoss
from .loss import LseHKRMulticlassLoss
from .loss import KRLoss
from .loss import KRMulticlassLoss
from .loss import NegKRLoss
from .loss import TauCrossEntropyLoss
from .loss import TauBCEWithLogitsLoss
from .loss import CategoricalHingeLoss
from .module import LipschitzModule
from .module import Sequential
from .module import vanilla_model
from .pooling import ScaledAdaptiveAvgPool2d
from .pooling import ScaledAvgPool2d
from .pooling import ScaledL2NormPool2d
from .pooling import ScaledAdaptativeL2NormPool2d
from .upsampling import InvertibleUpSampling
from .normalization import LayerCentering
from .normalization import BatchCentering
from .unconstrained import PadConv2d
from .unconstrained import PadConv1d
from .residual import LipResidual



# -*- coding: utf-8 -*-
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# Copyright IRT Antoine de Saint Exupéry et Université Paul Sabatier Toulouse III - All
# rights reserved. DEEL is a research program operated by IVADO, IRT Saint Exupéry,
# CRIAQ and ANITI - https://www.deel.ai/
# =====================================================================================
import warnings
from typing import Optional
import torch
from .. import functional as F
from torch.nn import CrossEntropyLoss, BCEWithLogitsLoss
from torch import Tensor


class KRLoss(torch.nn.Module):
    def __init__(self, multi_gpu=False, reduction: str = "mean", true_values=None):
        r"""
        Loss that estimates the Wasserstein-1 distance using the Kantorovich-Rubinstein
        duality.
        The Kantorovich-Rubinstein duality is formulated as following:

        $$
        W_1(\mu, \nu) =
        \sup_{f \in Lip_1(\Omega)} \underset{\textbf{x} \sim \mu}{\mathbb{E}}
        \left[f(\textbf{x} )\right] -
        \underset{\textbf{x}  \sim \nu}{\mathbb{E}} \left[f(\textbf{x} )\right]
        $$

        Where mu and nu stands for the two distributions, the distribution where the
        label is 1 and the rest.

        Note that `input` and `target` must be of rank 2: (batch_size, 1) or
        (batch_size, C) for multilabel classification (with C categories).
        `target` accepts label values in (0, 1), (-1, 1), or pre-processed with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Using a multi-GPU/TPU strategy requires to set `multi_gpu` to True and to
        pre-process the labels `target` with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Args:
            multi_gpu (bool): set to True when running on multi-GPU/TPU
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
            true_values: depreciated.
        """
        super().__init__()
        self.reduction = reduction
        self.multi_gpu = multi_gpu
        assert (
            true_values is None
        ), "depreciated true_values should be None (use target>0)"

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.kr_loss(input, target, multi_gpu=self.multi_gpu)
        return F.apply_reduction(loss_batch, self.reduction)


class NegKRLoss(torch.nn.Module):
    def __init__(self, multi_gpu=False, reduction: str = "mean", true_values=None):
        """
        Loss that estimates the negative of the Wasserstein-1 distance using
        the Kantorovich-Rubinstein duality. See `KRLoss` for more details.

        Args:
            multi_gpu (bool): set to True when running on multi-GPU/TPU
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
            true_values: depreciated.
        """
        super().__init__()
        self.reduction = reduction
        self.multi_gpu = multi_gpu
        assert (
            true_values is None
        ), "depreciated true_values should be None (use target>0)"

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.neg_kr_loss(input, target, multi_gpu=self.multi_gpu)
        return F.apply_reduction(loss_batch, self.reduction)


class HingeMarginLoss(torch.nn.Module):
    def __init__(self, min_margin: float = 1.0, reduction: str = "mean"):
        """
        Hinge margin loss.

        Args:
            min_margin: The minimal margin to enforce.
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        self.reduction = reduction
        self.min_margin = min_margin

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.hinge_margin_loss(input, target, self.min_margin)
        return F.apply_reduction(loss_batch, self.reduction)


class HKRLoss(torch.nn.Module):
    def __init__(
        self,
        alpha: float,
        min_margin: float = 1.0,
        multi_gpu=False,
        reduction: str = "mean",
        true_values=None,
    ):
        """
        Loss that estimates the Wasserstein-1 distance using the Kantorovich-Rubinstein
        duality with a hinge regularization.

        [1] M. Serrurier, F. Mamalet, et al. «Achieving robustness in classification
        using optimal transport with hinge regularization», 2021.

        Note that `input` and `target` must be of rank 2: (batch_size, 1) or
        (batch_size, C) for multilabel classification (with C categories).
        `target` accepts label values in (0, 1), (-1, 1), or pre-processed with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Using a multi-GPU/TPU strategy requires to set `multi_gpu` to True and to
        pre-process the labels `target` with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        the regularization factor `alpha` is a value between 0 and 1. It controls the
        trade-off between the hinge and the KR loss. When `alpha` is 0, the loss is
        equivalent to the KR loss, and when `alpha` is 1, the loss is equivalent to the
        hinge loss.

        Args:
            alpha: Regularization factor ([0,1]) between the hinge and the KR loss.
            min_margin: Minimal margin for the hinge loss.
            multi_gpu (bool): set to True when running on multi-GPU/TPU
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
            true_values: depreciated.
        """
        super().__init__()
        self.reduction = reduction
        self.multi_gpu = multi_gpu
        if (alpha >= 0) and (alpha <= 1):
            self.alpha = alpha
        else:
            warnings.warn(f"Depreciated alpha should be between 0 and 1 replaced by \
                    {alpha/(alpha+1.0)}")
            self.alpha = alpha / (alpha + 1.0)
        self.min_margin = min_margin
        assert (
            true_values is None
        ), "depreciated true_values should be None (use target>0)"

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.hkr_loss(
            input, target, self.alpha, self.min_margin, multi_gpu=self.multi_gpu
        )
        return F.apply_reduction(loss_batch, self.reduction)


class KRMulticlassLoss(torch.nn.Module):
    def __init__(self, multi_gpu=False, reduction: str = "mean"):
        r"""
        Loss to estimate average of Wasserstein-1 distance using Kantorovich-Rubinstein
        duality over outputs. In this multiclass setup, the KR term is computed for each
        class and then averaged.

        Note that`target` should be one-hot encoded or pre-processed with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Using a multi-GPU/TPU strategy requires to set `multi_gpu` to True and to
        pre-process the labels `target` with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Args:
            multi_gpu (bool): set to True when running on multi-GPU/TPU
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        self.reduction = reduction
        self.multi_gpu = multi_gpu

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.kr_loss(input, target, multi_gpu=self.multi_gpu)
        return F.apply_reduction(loss_batch, self.reduction)


class HingeMulticlassLoss(torch.nn.Module):
    def __init__(self, min_margin: float = 1.0, reduction: str = "mean"):
        r"""
        Loss to estimate the Hinge loss in a multiclass setup. It computes the
        element-wise hinge term. Note that this formulation differs from the
        one commonly found in tensorflow/pytorch (with maximise the difference
        between the two largest logits). This formulation is consistent with the
        binary classification loss used in a multiclass fashion.

        Args:
            min_margin: The minimal margin to enforce.
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        self.min_margin = min_margin
        self.reduction = reduction

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.hinge_multiclass_loss(input, target, self.min_margin)
        return F.apply_reduction(loss_batch, self.reduction)


class HKRMulticlassLoss(torch.nn.Module):
    def __init__(
        self,
        alpha: float,
        min_margin: float = 1.0,
        multi_gpu=False,
        reduction: str = "mean",
    ):
        """
        Loss that estimates the Wasserstein-1 distance using the Kantorovich-Rubinstein
        duality with a hinge regularization.

        [1] M. Serrurier, F. Mamalet, et al. «Achieving robustness in classification
        using optimal transport with hinge regularization», 2021.

        Note that`target` should be one-hot encoded or pre-processed with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        Using a multi-GPU/TPU strategy requires to set `multi_gpu` to True and to
        pre-process the labels `target` with the
        `deel.torchlip.functional.process_labels_for_multi_gpu()` function.

        the regularization factor `alpha` is a value between 0 and 1. It controls the
        trade-off between the hinge and the KR loss. When `alpha` is 0, the loss is
        equivalent to the KR loss, and when `alpha` is 1, the loss is equivalent to the
        hinge loss.

        Args:
            alpha: Regularization factor ([0,1]) between the hinge and the KR loss.
            min_margin: Minimal margin for the hinge loss.
            multi_gpu (bool): set to True when running on multi-GPU/TPU
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        if (alpha >= 0) and (alpha <= 1):
            self.alpha = alpha
        else:
            warnings.warn(f"Depreciated alpha should be between 0 and 1 replaced by \
                    {alpha/(alpha+1.0)}")
            self.alpha = alpha / (alpha + 1.0)
        self.min_margin = min_margin
        self.reduction = reduction
        self.multi_gpu = multi_gpu

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        loss_batch = F.hkr_multiclass_loss(
            input, target, self.alpha, self.min_margin, multi_gpu=self.multi_gpu
        )
        return F.apply_reduction(loss_batch, self.reduction)


class SoftHKRMulticlassLoss(torch.nn.Module):
    def __init__(
        self,
        alpha=0.9,
        min_margin=1.0,
        alpha_mean=0.99,
        temperature=1.0,
        reduction: str = "mean",
    ):
        """
        The multiclass version of HKR with softmax. This is done by computing
        the HKR term over each class and averaging the results.

        [2] M. Serrurier, F. Mamalet, T. Fel et al. "On the explainable properties
        of 1-Lipschitz Neural Networks: An Optimal Transport Perspective.", 2024

        Note that`target` should be one-hot encoded, +/-1 values.

        the regularization factor `alpha` is a value between 0 and 1. It controls the
        trade-off between the hinge and the KR loss. When `alpha` is 0, the loss is
        equivalent to the KR loss, and when `alpha` is 1, the loss is equivalent to the
        hinge loss.

        Args:
            alpha (float): regularization factor (0 <= alpha <= 1),
            min_margin (float): margin to enforce.
            alpha_mean (float): geometric mean factor
            temperature (float): factor for softmax  temperature
                (higher value increases the weight of the highest non y_true logits)
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        if (alpha >= 0) and (alpha <= 1):
            self.alpha = torch.tensor(alpha, dtype=torch.float32)
        else:
            warnings.warn(f"Depreciated alpha should be between 0 and 1 replaced by \
                    {alpha/(alpha+1.0)}")
            self.alpha = torch.tensor(alpha / (alpha + 1.0), dtype=torch.float32)
        self.min_margin_v = min_margin
        self.alpha_mean = alpha_mean

        self.current_mean = torch.tensor((self.min_margin_v,), dtype=torch.float32)
        """    constraint=lambda x: torch.clamp(x, 0.005, 1000),
            name="current_mean",
        )"""

        self.temperature = temperature * self.min_margin_v
        if alpha == 1.0:  # alpha = 1.0 => hinge only
            self.fct = self.multiclass_hinge_soft
        else:
            if alpha == 0.0:  # alpha = 0.0 => KR only
                self.fct = self.kr_soft
            else:
                self.fct = self.hkr
        self.reduction = reduction

        super(SoftHKRMulticlassLoss, self).__init__()

    def clamp_current_mean(self, x):
        return torch.clamp(x, 0.005, 1000)

    def _update_mean(self, y_pred):
        self.current_mean = self.current_mean.to(y_pred.device)
        current_global_mean = torch.mean(torch.abs(y_pred)).to(
            dtype=self.current_mean.dtype
        )
        current_global_mean = (
            self.alpha_mean * self.current_mean
            + (1 - self.alpha_mean) * current_global_mean
        )
        self.current_mean = self.clamp_current_mean(current_global_mean).detach()
        total_mean = current_global_mean
        total_mean = torch.clamp(total_mean, self.min_margin_v, 20000)
        return total_mean

    def computeTemperatureSoftMax(self, y_true, y_pred):
        total_mean = self._update_mean(y_pred)
        current_temperature = (
            torch.clamp(self.temperature / total_mean, 0.005, 250)
            .to(dtype=y_pred.dtype)
            .detach()
        )
        min_value = torch.tensor(torch.finfo(torch.float32).min, dtype=y_pred.dtype).to(
            device=y_pred.device
        )
        opposite_values = torch.where(
            y_true > 0, min_value, current_temperature * y_pred
        )
        F_soft_KR = torch.softmax(opposite_values, dim=-1)
        one_value = torch.tensor(1.0, dtype=F_soft_KR.dtype).to(device=y_pred.device)
        F_soft_KR = torch.where(y_true > 0, one_value, F_soft_KR)
        return F_soft_KR

    def signed_y_pred(self, y_true, y_pred):
        """Return for each item sign(y_true)*y_pred."""
        sign_y_true = torch.where(y_true > 0, 1, -1)  # switch to +/-1
        sign_y_true = sign_y_true.to(dtype=y_pred.dtype)
        return y_pred * sign_y_true

    def multiclass_hinge_preproc(self, signed_y_pred, min_margin):
        """From multiclass_hinge(y_true, y_pred, min_margin)
        simplified to use precalculated signed_y_pred"""
        # compute the elementwise hinge term
        hinge = torch.nn.functional.relu(min_margin / 2.0 - signed_y_pred)
        return hinge

    def multiclass_hinge_soft_preproc(self, signed_y_pred, F_soft_KR):
        hinge = self.multiclass_hinge_preproc(signed_y_pred, self.min_margin_v)
        b = hinge * F_soft_KR
        b = torch.sum(b, axis=-1)
        return b

    def multiclass_hinge_soft(self, y_true, y_pred):
        F_soft_KR = self.computeTemperatureSoftMax(y_true, y_pred)
        signed_y_pred = self.signed_y_pred(y_true, y_pred)
        return self.multiclass_hinge_soft_preproc(signed_y_pred, F_soft_KR)

    def kr_soft_preproc(self, signed_y_pred, F_soft_KR):
        kr = -signed_y_pred
        a = kr * F_soft_KR
        a = torch.sum(a, axis=-1)
        return a

    def kr_soft(self, y_true, y_pred):
        F_soft_KR = self.computeTemperatureSoftMax(y_true, y_pred)
        signed_y_pred = self.signed_y_pred(y_true, y_pred)
        return self.kr_soft_preproc(signed_y_pred, F_soft_KR)

    def hkr(self, y_true, y_pred):
        F_soft_KR = self.computeTemperatureSoftMax(y_true, y_pred)
        signed_y_pred = self.signed_y_pred(y_true, y_pred)

        loss_softkr = self.kr_soft_preproc(signed_y_pred, F_soft_KR)

        loss_softhinge = self.multiclass_hinge_soft_preproc(signed_y_pred, F_soft_KR)
        return (1 - self.alpha) * loss_softkr + self.alpha * loss_softhinge

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if not (isinstance(input, torch.Tensor)):  # required for dtype.max
            input = torch.Tensor(input, dtype=input.dtype)
        if not (isinstance(target, torch.Tensor)):
            target = torch.Tensor(target, dtype=input.dtype)
        loss_batch = self.fct(target, input)
        return F.apply_reduction(loss_batch, self.reduction)


class LseHKRMulticlassLoss(torch.nn.Module):
    def __init__(
        self,
        alpha: float = 1.0,
        temperature: float = 1.0,
        penalty=1.0,  # max <logsumpexp< max+penalty*margin
        min_margin=1,
        reduction: str = "mean",
    ):
        """
        Loss that estimates the Wasserstein-1 distance using the Kantorovich-Rubinstein
        duality with a hinge regularization and logsumexp summary.
        Args:
            alpha: Regularization factor between the hinge and the KR loss
                (0 <= alpha <= 1).
            temperature (float): temperature factor applied on logits
                (both in KR and Hinge)
            penalty (float): penalty factor for the logsumexp summary
                max <logsumpexp< max+penalty*margin (default is 1.0).
            min_margin: Minimal margin for the hinge loss.
            reduction: type of reduction applied to the output. possible values are
                   'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        if (alpha >= 0) and (alpha <= 1):
            self.alpha = torch.tensor(alpha, dtype=torch.float32)
        else:
            warnings.warn(f"Depreciated alpha should be between 0 and 1 replaced by \
                    {alpha/(alpha+1.0)}")
            self.alpha = torch.tensor(alpha / (alpha + 1.0), dtype=torch.float32)
        self.penalty = penalty
        self.temperature = temperature
        self.margin = min_margin
        if alpha == 1.0:  # alpha = 1.0 => hinge only
            self.fct = self.lse_hinge
        else:
            if alpha == 0.0:  # alpha = 0.0 => KR only
                self.fct = self.lse_kr
            else:
                self.fct = self.lse_hkr
        self.reduction = reduction

    def get_positive(self, y_pred, y_true):
        return y_pred[y_true > 0]

    def compute_lse_neg(self, y_pred, y_true, min_margin):
        neg = torch.where(y_true > 0, -float("inf"), y_pred)
        nb_bins = y_pred.new_tensor(y_pred.size(1) - 1)
        nb_bins = torch.log(nb_bins)
        t = nb_bins / (min_margin * self.penalty / 2.0)  # margin = min_margin/2.0
        lse_neg = (
            1 / t * torch.logsumexp(t * neg, dim=1)
        )  # max <neg_soft< max+penalty*margin
        return lse_neg

    def lse_hinge_preproc(self, pos, lse_neg, min_margin):
        # compute the elementwise hinge term
        hinge_pos = torch.nn.functional.relu(min_margin / 2.0 - pos)
        hinge_neg = torch.nn.functional.relu(min_margin / 2.0 + lse_neg)
        return hinge_pos + hinge_neg

    def lse_hinge(self, y_pred, y_true):
        pos = self.get_positive(y_pred, y_true)
        lse_neg = self.compute_lse_neg(y_pred, y_true, self.margin)
        hinge = self.lse_hinge_preproc(pos, lse_neg, self.margin)
        return hinge

    def lse_kr_preproc(self, pos, lse_neg):
        return pos - lse_neg

    def lse_kr(self, y_pred, y_true):
        pos = self.get_positive(y_pred, y_true)
        lse_neg = self.compute_lse_neg(y_pred, y_true, self.margin)
        kr = -self.lse_kr_preproc(pos, lse_neg)
        return kr

    def lse_hkr(self, y_pred, y_true):
        pos = self.get_positive(y_pred, y_true)
        lse_neg = self.compute_lse_neg(y_pred, y_true, self.margin)

        lse_kr = -self.lse_kr_preproc(pos, lse_neg)
        lse_hinge = self.lse_hinge_preproc(pos, lse_neg, self.margin)
        return (1 - self.alpha) * lse_kr + self.alpha * lse_hinge

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        if not (isinstance(input, torch.Tensor)):  # required for dtype.max
            input = torch.Tensor(input, dtype=input.dtype)
        if not (isinstance(target, torch.Tensor)):
            target = torch.Tensor(target, dtype=input.dtype)
        input = input * self.temperature
        loss_batch = self.fct(input, target)
        return F.apply_reduction(loss_batch, self.reduction)


class TauCrossEntropyLoss(CrossEntropyLoss):
    def __init__(
        self,
        tau: float,
        weight: Optional[Tensor] = None,
        size_average=None,
        ignore_index: int = -100,
        reduce=None,
        reduction: str = "mean",
        label_smoothing: float = 0.0,
    ) -> None:
        """
        The loss add a temperature (tau) factor to the CrossEntropyLoss
        CrossEntropyLoss(tau * input, target).

        See `CrossEntropyLoss` for more details on arguments.

        Args:
            tau (float): factor for  temperature
        """

        super().__init__(
            weight=weight,
            size_average=size_average,
            ignore_index=ignore_index,
            reduce=reduce,
            reduction=reduction,
            label_smoothing=label_smoothing,
        )
        self.tau = tau

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if input.shape == target.shape:
            return super().forward(input * self.tau, target.to(torch.double)) / self.tau
        else:
            return super().forward(input * self.tau, target.to(torch.int64)) / self.tau


class TauBCEWithLogitsLoss(BCEWithLogitsLoss):
    def __init__(
        self,
        tau: float,
        weight: Optional[Tensor] = None,
        size_average=None,
        reduce=None,
        reduction: str = "mean",
        pos_weight=None,
    ) -> None:
        """
        The loss add a temperature (tau) factor to the BCEWithLogitsLoss
        BCEWithLogitsLoss(tau * input, target).

        See `BCEWithLogitsLoss` for more details on arguments.

        Args:
            tau (float): factor for  temperature
        """
        super().__init__(
            weight=weight,
            size_average=size_average,
            reduce=reduce,
            reduction=reduction,
            pos_weight=pos_weight,
        )
        self.tau = tau

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        true_target = torch.where(target > 0, 1.0, 0.0).to(input.dtype)
        return super().forward(input * self.tau, true_target) / self.tau


class CategoricalHingeLoss(torch.nn.Module):
    def __init__(
        self,
        min_margin: float = 1.0,
        reduction: str = "mean",
    ):
        """
        This implementation is sligthly different from the pytorch MultiMarginLoss.

        `target` and `input` must be of shape (batch_size, # classes).
        Note that `target` should be one-hot encoded, +/-1 values.
        :math:`\\text{ReLU}(\\text{min\\_margin} - (\\text{input}[\\text{target}>0]
        - \\text{max}(\\text{input}[\\text{target}<=0])))`
        is computed element-wise and averaged over the batch.

        Args:
            min_margin (float): margin parameter.
            reduction: type of reduction applied to the output. possible values are
                'none' | 'mean' | 'sum' | 'auto'; default is 'mean' ('auto is 'mean')
        """
        super().__init__()
        self.min_margin = min_margin
        self.reduction = reduction

    def forward(self, input: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        mask = torch.where(target > 0, 1, 0).to(input.dtype)
        pos = torch.sum(mask * input, dim=-1)
        neg = torch.max(
            torch.where(target > 0, input - self.min_margin, input), dim=-1
        ).values
        loss_batch = torch.nn.functional.relu(self.min_margin - (pos - neg))
        return F.apply_reduction(loss_batch, self.reduction)
